#!/usr/bin/env dart

/// Watches filesystem for updates to auto-compile/auto-test
import 'dart:io';
import 'package:args/args.dart';
import 'package:ebisu/ebisu.dart';
import 'package:logging/logging.dart';
import 'package:path/path.dart' as path;
import 'package:quiver/iterables.dart';

// custom <additional imports>
// end <additional imports>
//! The parser for this script
ArgParser _parser;
//! The comment and usage associated with this script
void _usage() {
  print(r'''
Watches filesystem for updates to auto-compile/auto-test
''');
  print(_parser.getUsage());
}

//! Method to parse command line options.
//! The result is a map containing all options, including positional options
Map _parseArgs(List<String> args) {
  ArgResults argResults;
  Map result = {};
  List remaining = [];

  _parser = new ArgParser();
  try {
    /// Fill in expectations of the parser
    _parser.addFlag('help',
        help: r'''
Display this help screen
''',
        abbr: 'h',
        defaultsTo: false);

    _parser.addOption('ebisu-cpp-project',
        help: r'''
Path to ebisu cpp project
''',
        defaultsTo: null,
        allowMultiple: false,
        abbr: 'p',
        allowed: null);
    _parser.addOption('log-level',
        help: r'''
Select log level from:
[ all, config, fine, finer, finest, info, levels,
  off, severe, shout, warning ]

''',
        defaultsTo: null,
        allowMultiple: false,
        abbr: null,
        allowed: null);

    /// Parse the command line options (excluding the script)
    argResults = _parser.parse(args);
    if (argResults.wasParsed('help')) {
      _usage();
      exit(0);
    }
    result['ebisu-cpp-project'] = argResults['ebisu-cpp-project'];
    result['help'] = argResults['help'];
    result['log-level'] = argResults['log-level'];

    if (result['log-level'] != null) {
      const choices = const {
        'all': Level.ALL,
        'config': Level.CONFIG,
        'fine': Level.FINE,
        'finer': Level.FINER,
        'finest': Level.FINEST,
        'info': Level.INFO,
        'levels': Level.LEVELS,
        'off': Level.OFF,
        'severe': Level.SEVERE,
        'shout': Level.SHOUT,
        'warning': Level.WARNING
      };
      final selection = choices[result['log-level'].toLowerCase()];
      if (selection != null) Logger.root.level = selection;
    }

    return {'options': result, 'rest': argResults.rest};
  } catch (e) {
    _usage();
    throw e;
  }
}

final _logger = new Logger('cppWatcher');

class CppProject {
  Iterable cppPaths = [];

  /// Files that look like make files requiring rebuild
  Iterable makelikeFiles = [];

  // custom <class CppProject>

  CppProject(this.cppPaths, this.makelikeFiles);

  watch() {
    for (var cppPath in cppPaths) {
      new Directory(cppPath).watch().listen((var e) {
        switch (e.runtimeType) {
          case FileSystemModifyEvent:
            print(e);
        }
      });
    }
  }

  // end <class CppProject>

}

class CppProjectFinder {
  String rootPath;

  // custom <class CppProjectFinder>

  final ignoredDirs = new RegExp(
      r'/(?:.git|.pub|build_dir|install|cmakefiles|packages)$',
      caseSensitive: false);
  final cishFiles = new RegExp(r'c|cpp|h|hpp', caseSensitive: false);

  nextDir(d) => concat(new Directory(d)
      .listSync()
      .where((fse) =>
          (fse is Directory && !fse.path.contains(ignoredDirs)) || fse is File)
      .map((fse) => fse is Directory ? nextDir(fse.path) : [fse.path]));

  CppProjectFinder(this.rootPath) {
    final cppPaths = [];
    final makishFiles = [];
    final makishRe = new RegExp(r'(?:cmakelists|makefile)(?:\.(?:mak|mk))?',
        caseSensitive: false);
    final cppDirs = new Set();
    final makeishFiles = new Set();

    nextDir(rootPath).forEach((var file) {
      if (path.extension(file).contains(cishFiles))
        cppDirs.add(path.dirname(file));
      if (file.contains(makishRe)) makishFiles.add(file);
    });

    final cppProject = new CppProject(cppDirs, makishFiles);
    cppProject.watch();

    // print(brCompact(cppDirs));
    // print('Makish');
    // print(brCompact(makishFiles));
  }

  // end <class CppProjectFinder>

}

main(List<String> args) {
  Logger.root.onRecord.listen(
      (LogRecord r) => print("${r.loggerName} [${r.level}]:\t${r.message}"));
  Logger.root.level = Level.OFF;
  Map argResults = _parseArgs(args);
  Map options = argResults['options'];
  List positionals = argResults['rest'];
  // custom <cppWatcher main>

  new CppProjectFinder(options['ebisu-cpp-project']);

  // end <cppWatcher main>
}

// custom <cppWatcher global>
// end <cppWatcher global>
